/**
  User-level threads example, showing how threads can 
  be written to disk and restored.
  Orion Sky Lawlor, olawlor@acm.org, 2005/2/18 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ucontext.h> /* for makecontext/swapcontext routines */

/* Tiny little user-level thread library */
typedef void (*threadFn)(void);

/* Create a new user-level thread to run this function */
void makeThread(ucontext_t *T,threadFn fn) {
	getcontext(T);
	/* Allocate a stack for the new thread */
	int stackLen=32*1024;
	char *stack=new char[stackLen];
	memset(stack,0,stackLen);
	printf("New thread stack is at %p\n",stack);
	T->uc_stack.ss_sp=stack;
	T->uc_stack.ss_size=stackLen;
	T->uc_stack.ss_flags=0;
	makecontext(T,fn,0);
}

/* Save this user-level thread to disk.
  Saves the thread's registers and stack, but *NO* heap data.
*/
void saveThread(ucontext_t *T,const char *fName) {
	FILE *f=fopen(fName,"wb");
	fwrite(T,1,sizeof(*T),f);
	fwrite(T->uc_stack.ss_sp,1,T->uc_stack.ss_size,f);
	fclose(f);
	printf("Checkpoint written.\n");
}

/* Read this user-level thread from this disk file.
  Restores the thread's registers and stack, but *NO* heap data.
*/
void restoreThread(ucontext_t *T,FILE *f) {
	fread(T,1,sizeof(*T),f);
	fread(T->uc_stack.ss_sp,1,T->uc_stack.ss_size,f);
	fclose(f);
	printf("Checkpoint read.\n");
}


/* Context structure for a user-level thread. */
ucontext_t A;

int nIter=1000;

void runA(void) {
	for (int n=0;n<nIter;n++) {
		char buf[100];
		sprintf(buf,"Yup.  This here's the stack...  I'm on iteration %d.\n",n);
		for (int i=0;i<10*1000*1000;i++) {}
		printf("A(%d)\n",n);
		
		if (n%10==0) { /* Save our thread context to disk. */
			printf("Checkpointing: about to call getcontext\n");
			getcontext(&A);
			printf("Checkpointing: back from getcontext\n");
			saveThread(&A,"saved.ctx");
		}
	}
}

int main() {
	makeThread(&A,runA);
	FILE *f=fopen("saved.ctx","rb"); // Load from checkpoint
	if (f) restoreThread(&A,f);
	setcontext(&A); /* Start running thread A first */
	return 0;
}

